#include <stdlib.h>
#include <math.h>


/*	Find the bin counts for the data array x(n) for the ASH 
	estimator.  nbin bins are formed over the interval [a,b).
	Bin counts are returned in the array "nc" - and the 
	number of points outside [a,b) in "nskip"		*/

void bin1(x,n,ab,nbin,nc,nskip)

int *n, *nbin, *nc, *nskip;
double *x, *ab; 

{
    int i,k;
    double a,b,d;

    *nskip=0;
    a=ab[0];
    b=ab[1];
    for (i=0;i<(*nbin);i++) nc[i]=0;

    d=(b-a)/((double)(*nbin));

    for (i=0;i<(*n);i++) {
	k=((int)((x[i]-a)/d+1.0));
	if ((k>=1)&&(k<=((double)(*nbin)))) nc[k-1]++;
	else (*nskip)++;	
    } 
}

/*	Compute the ASH density estimate in 1 dimension using the
	Quartic (biweight) kernel.  Average of m shifted histograms.
	Bin counts in array nc from routine bin1.  nbin bins are formed
	over the interval [a,b).  ASH estimates are returned in the
	array "f".  The ASH should be plotted along the grid of
	a+(d/2),...,b-(d/2), where d=(b-a)/nbin.  Also, if nskip
	was non-zero, ASH estimates are incorrect near the boundary.
	Should leave m empty bins on each end of array "nc" so f is OK.  */
	
void ash1(m,nc,nbin,ab,kopt,t,f,w,ier)

int *m, *nc, *nbin, *ier;
double *ab, *kopt, *w, *f, *t;

{
    int n,i,k,l1,l2,d,mm1;
    double a,b,cons,delta,h,c,q1,xm;
 
    *ier=0; 
    a=ab[0];
    b=ab[1];
    n=0;
    mm1=*m-1;
    xm=((double)(*m));

    w[0]=1.0;
    cons=1.0;

    for (i=1;i<=mm1;i++) {
	q1=((double)(i))/xm;
	q1=fabs(q1);
	q1=pow(q1,kopt[0]);
	q1=1.0-q1;
	q1=pow(q1,kopt[1]);
       	w[i]=q1;
       	cons+=2.0*w[i];
    }
    cons=((double)(*m))/cons;
    for (i=1;i<=(*m);i++) w[i-1]*=cons;
    for (i=1;i<=(*m-1);i++) if((nc[i-1]+nc[*nbin+2-i])>0) *ier=1;
    delta=(b-a)/((double)(*nbin));
    h=xm*delta;
    for (i=1;i<=(*nbin);i++) {
	t[i-1]=a+(((double)(i))-0.5)*delta;
	f[i-1]=0.0;
	n+=nc[i-1];
    } 
    for (i=1;i<=(*nbin);i++) {
	if (nc[i-1]>0) {
	    c=((double)(nc[i-1]))/(h*((double)(n)));
	    l1=1;
	    if ((i-(*m)+1)>1) l1=i-(*m)+1;
	    l2=*nbin;
	    if ((i+(*m)-1)<(*nbin)) l2=i+(*m)-1;
	    for (k=l1;k<=l2;k++) {
		if (k>=i) d=k-i;
            	else d=i-k;
		f[k-1]+=c*w[d]; 
	    }
	}
    }
}

void bin2(x,n,ab,nbin,nc,nskip)

int *n,*nc,*nbin,*nskip;
double *x,*ab;
{
    int j,i,kx,ky,nbin1,nbin2;
    double ax, ay, bx, by, dx, dy;

    *nskip=0;
    ax=ab[0];
    bx=ab[2];
    ay=ab[1];
    by=ab[3]; 

    nbin1=nbin[0];
    nbin2=nbin[1];

    for (j=1;j<=nbin2;j++) for (i=1;i<=nbin1;i++) nc[(j-1)*nbin1+i-1]=0;
    dx=(bx-ax)/((double)(nbin1));
    dy=(by-ay)/((double)(nbin2));

    for (i=1;i<=(*n);i++) {
	kx=((int)((x[i-1]-ax)/dx+1.0));
	ky=((int)((x[(*n)+i-1]-ay)/dy+1.0));
	if ((kx>=1)&&(kx<=nbin1)&&(ky>=1)&&(ky<=nbin2)) nc[(ky-1)*nbin1+kx-1]++;
	else (*nskip)++;
    }
}

void ash2(m,nc,nbinx,nbiny,ab,kopt,f,w,ier)

int *m,*nc,*nbinx,*nbiny,*ier;
double *ab, *kopt, *f, *w;

{
    int i,j,n,ncheck,kx,ky,x1,x2,y1,y2;
    int mx,my,mxm1,mym1,xm,ym,w1,w2;
    int bin1,bin2;
    double ax,bx,ay,by,consx,consy;
    double dx,dy,hx,hy,c;

    *ier=0;
    ax=ab[0];
    bx=ab[2];
    ay=ab[1];
    by=ab[3];

    mx=m[0];
    my=m[1];
    mxm1=mx-1;
    mym1=my-1;
    xm=mx;
    ym=my;
    f[0]=1.0;
    f[1]=1.0;
   
    consx=1.0;
    consy=1.0;
    bin1=*nbinx;
    bin2=*nbiny;

    for (i=1;i<=mxm1;i++) { 
	f[bin1*i]=pow((1.0-pow(fabs(((double)(i))/
		((double)(xm))),kopt[0])),kopt[1]); 
	consx+=2.0*f[bin1*i];
    }
    consx=((double)(mx))/consx;

    for (i=1;i<=mym1;i++) { 
	f[bin1*i+1]=pow((1.0-pow(fabs(((double)(i))/
		((double)(ym))),kopt[0])),kopt[1]);
	consy+=2.0*f[bin1*i+1];
    }
    consy=((double)(my))/consy;

    for (j=1;j<=my;j++) for (i=1;i<=mx;i++) 
	w[(j-1)*bin1+i-1]=(consx*(f[(i-1)*bin1]))*(consy*(f[(j-1)*bin1+1]));

    n=0;
    for (j=1;j<=bin2;j++) for (i=1;i<=bin1;i++) {
	f[(j-1)*bin1+i-1]=0.0;
	n+=nc[(j-1)*bin1+i-1];
    }

    ncheck=0;
    for (j=my;j<=(bin2-my+1);j++)
	for (i=mx;i<=(bin1-mx+1);i++) {
		ncheck+=nc[(j-1)*bin1+i-1];
	}

    if (ncheck!=n) *ier=1;

    dx=(bx-ax)/((double)(bin1));
    dy=(by-ay)/((double)(bin2)); 
    hx=((double)(mx))*dx;
    hy=((double)(my))*dy;

    for (j=1;j<=bin2;j++) for (i=1;i<=bin1;i++) 
      if (nc[(j-1)*bin1+i-1]>0) {
	c=((double)(nc[(j-1)*bin1+i-1]))/(hx*hy*((double)(n)));
	y1=1;
	if ((j-mym1)>1) y1=j-mym1;
  	y2=bin2;
	if ((j+mym1)<bin2) y2=j+mym1;
	x1=1;
	if ((i-mxm1)>1) x1=i-mxm1;
	x2=bin1;
	if ((i+mxm1)<bin1) x2=i+mxm1;
	for (ky=y1;ky<=y2;ky++) for (kx=x1;kx<=x2;kx++) {
	    if (kx>=i) w1=kx-i;
	    else w1=i-kx;
	    if (ky>=j) w2=ky-j;
	    else w2=j-ky;
	    f[(ky-1)*bin1+kx-1]+=c*w[w2*bin1+w1];
	}
    } 
}
