/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Demo Program for the Density Grand Tour...		 */
/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
#define pi 3.14159265
#define MAXVERTS 10000
#define MAXBINS 201
#define RUNLENGTH 1000
#define MAXDIM 10
#define MAXGROUPS 5
#define OUT 1

#include "forms.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "ash.c"
#include "contour.c"
#include "ran2.c"
#include "dgtX1d.c"
#include "loading.c"

main(argc, argv)
    int argc;
    char **argv;
{
    FD_dgtX1d *dgtform;
    FD_loading *loading;

    FL_OBJECT *obj; 
    int n, p, nt, i, j, nbin, nskip, m, ier;
    int pp, tm, ng;
    int loc;
    int *nc;
    int k, l, nv, ny, rel;
    int beg_nv;
    int ii, jj, *G;
    int *Group_Numbers;
    int Temp_Int=0;
    int *number;
    int step;	      /* OVERALL STEP of the DGT   */
    int prev_step;
    int INTERACTIVE=0;/* 0 = COMPUTER, 0 = USER	   */ 
    int DIRECTION=1;  /* 1 = FORWARD, 0 = BACKWARD */
    int QUIT=0;
    int DRAWWTS=0;
    int DRAWPTS=0;
    int BRUSH=0;
    int BINS;
    int SMPAR;
    int NEWSTEP;
    int STEPDIFF;
    int USERACT;

    long *idum; 
    long start=-1;

    double ff[MAXBINS][MAXGROUPS];
    double hmult=2.0;
    double *newloc;
    double ll,ul;
    double t, ab[2], kopt[2], hfac;
    double *X, *J;
    double *Temp_X;
    double lambda_store[10000];
    double lambda[10];
    double *min, *max;
    double maxlength, clength;
    double *f,*w;
    double *wts;
    double *v, *xx, *yy, *currv;
    double dx, dy, dt, seed;
    double maxf;
    double *wtmatrix;
    double *Temp_Double;
    double zspace;
    char word[50], filename[50];
    char *input;
    FILE *fp;

    void rotated();
    void newlambda();
    void multmat();
    void outquad();
    void startquad();
    void endquad();

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*	Input the data from the file "demo.info"		 */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */

    if (!(fp=fopen("demo1d.info","r")))  {
            printf("Could not open file \"demo.info\"\n");
            fclose(fp);
            exit(1); }

/* -==-==-==-=-==-==-==-==-==-== */
/* Get the filename              */
/* -==-==-==-=-==-==-==-==-==-== */
    fscanf(fp,"%s",filename);

    fl_initialize(&argc, argv, "Density Grand Tour", 0, 0);

    loading=create_form_loading(filename);
    fl_show_form(loading->loading,FL_PLACE_HOTSPOT,FL_TRANSIENT,"Loading...");

/* -==-==-==-=-==-==-==-==-==-== */
/* Get the number of data points */
/* -==-==-==-=-==-==-==-==-==-== */
    fscanf(fp,"%s",word); /* remove 'DIMENSIONS:' */
    fscanf(fp,"%s",word);
    n=atoi(word);
/* -==-==-==-=-==-==-==-==-==-== */
/* Get the number of dimensions  */
/* -==-==-==-=-==-==-==-==-==-== */
    fscanf(fp,"%s",word);
    p=atoi(word);
    pp=(int)((pow(((double)(p)),2.0)-((double)(p)))/2.0);

/* -==-==-==-=-==-==-==-==-==-== */
/* Get the number of groups	 */
/* -==-==-==-=-==-==-==-==-==-== */
    fscanf(fp,"%s",word); /* remove 'GROUPS:' */
    fscanf(fp,"%s",word);
    ng=atoi(word);

/* -==-==-==-=-==-==-==-==-==-== */
/* Get the number of contours    */
/* -==-==-==-=-==-==-==-==-==-== */
    fscanf(fp,"%s",word); /* remove 'CONTOURS:" */
    fscanf(fp,"%s",word);
    nv=atoi(word);

/* -==-==-==-=-==-==-==-==-==-== */
/* Allocate space for the levels */
/*  v : the relative height (0-1)*/
/*  currv : v * maxheight        */
/* -==-==-==-=-==-==-==-==-==-== */
    v=(double *) calloc(nv,sizeof(double));
    currv=(double *) calloc(nv,sizeof(double));

/* -==-==-==-=-==-==-==-==-==-== */
/* Get the contour levels        */
/* -==-==-==-=-==-==-==-==-==-== */
    for (i=0;i<nv;i++) {
	fscanf(fp,"%s",word);
	v[i]=atof(word);
    } 

/* -==-==-==-=-==-==-==-==-==-== */
/* Get the smoothing parameter   */
/* -==-==-==-=-==-==-==-==-==-== */
    fscanf(fp,"%s",word); /* remove 'M:" */
    fscanf(fp,"%s",word);
    m=atoi(word);

/* -==-==-==-=-==-==-==-==-==-== */
/* Get the number of bins        */
/* -==-==-==-=-==-==-==-==-==-== */
    fscanf(fp,"%s",word); /* remove 'bins:' */
    fscanf(fp,"%s",word);
    nbin=atoi(word);

/* -==-==-==-=-==-==-==-==-==-== */
/* Get the kernel parameter      */
/* -==-==-==-=-==-==-==-==-==-== */
    fscanf(fp,"%s",word); /* remove 'kopt:' */
    fscanf(fp,"%s",word);
    kopt[0]=atof(word);
    fscanf(fp,"%s",word);
    kopt[1]=atof(word);

    fscanf(fp,"%s",word);
    fscanf(fp,"%s",word);
    dt=atof(word);

    ul=1.0/dt;
    ll=.5*ul;

    fscanf(fp,"%s",word);
    fscanf(fp,"%s",word);
    seed=atof(word);

    fclose(fp);

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Initialize storage variables                             */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
    nc=(int *) calloc(1000,sizeof(int));
    Group_Numbers=(int *) calloc(ng,sizeof(int));
    for (i=1;i<=ng;i++) Group_Numbers[i-1]=0;
    X=(double *) calloc((n*p),sizeof(double));
    Temp_X=(double *) calloc((n*p),sizeof(double));
    Temp_Double=(double *) calloc(p,sizeof(double));
    J=(double *) calloc(n,sizeof(double));
    G=(int *) calloc(n,sizeof(int));
    f=(double *) calloc(1000,sizeof(double));
    w=(double *) calloc(1000,sizeof(double));
    xx=(double *) calloc(1000,sizeof(double));
    min=(double *) calloc(p,sizeof(double));
    max=(double *) calloc(p,sizeof(double));
    wts=(double *) calloc(p,sizeof(double));
    wtmatrix=(double *) calloc((p*p),sizeof(double));

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Initialize the 'PIPE' to geomview                        */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
    printf("(normalization World all)\n");
    printf("(geometry dgt1d { : foo })\n");
    fflush(stdout);

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Initialize the Random Seed                               */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
    srand(seed);
    start*=rand();
    idum=&start;

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Input the data from the file *filename                   */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
    if (!(fp=fopen(filename,"r")))  {
            printf("Could not open file \"%s\"\n",filename);
            fclose(fp);
            exit(1); }

    for (i=1;i<=n;i++) {
	for (j=1;j<=p;j++) {
            fscanf(fp,"%s",word);
            X[(j-1)*n+(i-1)]=((double)(atof(word)));
	    if (i==1) {
            	min[j-1]=X[(j-1)*n+(i-1)];
            	max[j-1]=X[(j-1)*n+(i-1)];
            } else {
            	if (X[(j-1)*n+(i-1)]<min[j-1]) min[j-1]=X[(j-1)*n+(i-1)];
            	if (X[(j-1)*n+(i-1)]>max[j-1]) max[j-1]=X[(j-1)*n+(i-1)];
            }
	}
	fscanf(fp,"%s",word);
 	G[i-1]=atoi(word);	
	Group_Numbers[G[i-1]-1]++;
    }
    fclose(fp);

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Sort the data by groups                                  */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
    Temp_Int=0;
    for (i=1;i<n;i++)   /* Do we need to sort?  */
	if (G[i-1]>G[i]) {
	    Temp_Int++; 
	    break;
	}		/* Then sort....	*/
    if (Temp_Int!=0) for (i=1;i<n;i++) for (j=1;j<=(n-i);j++) {
	if (G[j-1]>G[j]) {      /* Swap */
	    for (k=1;k<=p;k++) Temp_Double[k-1]=X[(k-1)*n+(j-1)];
	    for (k=1;k<=p;k++) X[(k-1)*n+(j-1)]=X[(k-1)*n+j];
	    for (k=1;k<=p;k++) X[(k-1)*n+j]=Temp_Double[k-1];
	    Temp_Int=G[j-1];
	    G[j-1]=G[j];
	    G[j]=Temp_Int;
	}
    } 

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Rescale the data into [-1,1] by dimension                */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
    for (i=1;i<=n;i++) for (j=1;j<=p;j++) {
	X[(j-1)*n+(i-1)]=
	    2.0*(X[(j-1)*n+(i-1)]-min[p-1])/(max[p-1]-min[p-1])-1.0;
    }

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Shrink the data to fit in the unit hypersphere           */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
    maxlength=0.0;
    for (i=1;i<=n;i++) {
	clength=0.0;
	for (j=1;j<=p;j++) {
    	    clength+=pow(X[(j-1)*n+(i-1)],2.0);
        }
    	clength=sqrt(clength);
	if (clength>maxlength) maxlength=clength;
    }
    for (i=1;i<=n;i++) for (j=1;j<=p;j++) X[(j-1)*n+i-1]/=maxlength;

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Create the Jittering Vector in [0,.25]                   */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
    for (i=1;i<=n;i++) J[i-1]=ran1(idum)*.25;

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Create the bin edge vector                               */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
    ab[0]=-1.5;
    ab[1]=1.5;
    dx=(ab[1]-ab[0])/((double)(nbin));

    fl_hide_form(loading->loading);
    dgtform=create_form_dgtX1d(filename);
    fl_show_form(dgtform->dgtX1d,FL_PLACE_MOUSE,1,"Density Grand Tour");
    fl_set_counter_value(dgtform->SmParX, ((double)(m)));
    fl_set_counter_value(dgtform->binsX, ((double)(nbin)));

/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */
/*      Run the 1-d Density Grand Tour                           */
/* -==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==-==- */

    for (i=1;i<=p;i++) for (k=1;k<=p;k++) 
	if (i==k) wtmatrix[(i-1)*p+k-1]=1.0;
	else wtmatrix[(i-1)*p+k-1]=0.0;

    step=0;
    while (1==1) {
	nt=(int)(ll+ran1(idum)*(ul-ll)+1.0);
	newlambda(lambda,pp,idum);
	for (j=1;j<=nt;j++) {
	    step++;
	    for (k=0;k<pp;k++) lambda_store[(step-1)*pp+k]=lambda[k]; 
	    if (step>=RUNLENGTH) break;
	}
	if (step>=RUNLENGTH) break;
    }
    step=0;

    fl_do_forms(); 
    while (step <= RUNLENGTH) {

	/* -=-=-=-=-=-=-=-=-=- */
	/* Retrieve user input */ 	
	/* -=-=-=-=-=-=-=-=-=- */

	obj=fl_check_forms();
	BINS=((int) fl_get_counter_value(dgtform->binsX));
	SMPAR=((int) fl_get_counter_value(dgtform->SmParX));
	QUIT=fl_get_button(dgtform->quit);
	INTERACTIVE=fl_get_button(dgtform->frame_by_frame);
   	DRAWWTS=fl_get_button(dgtform->weights);
 	DRAWPTS=fl_get_button(dgtform->drawpts);
	BRUSH=fl_get_button(dgtform->brushX);

	while (((step==1)&&(DIRECTION==0)&&(INTERACTIVE==0))||
	    ((step==1000)&&(DIRECTION==1)&&(INTERACTIVE==0))) {
	    obj=fl_check_forms();
	    NEWSTEP=fl_get_counter_value(dgtform->frame);
	    if (NEWSTEP>step) DIRECTION=1;
	    else if (NEWSTEP<step) DIRECTION=0;
	    BINS=((int) fl_get_counter_value(dgtform->binsX));
            SMPAR=((int) fl_get_counter_value(dgtform->SmParX));
            QUIT=fl_get_button(dgtform->quit);
            INTERACTIVE=fl_get_button(dgtform->frame_by_frame);
            DRAWWTS=fl_get_button(dgtform->weights);
            DRAWPTS=fl_get_button(dgtform->drawpts);
            BRUSH=fl_get_button(dgtform->brushX);
	}

	/* -=-=-=-=-=-=-=-=-=- */
        /* Update parameters   */       
        /* -=-=-=-=-=-=-=-=-=- */

	if (QUIT==1) exit(1);

  	if (nbin != BINS) {
            nbin=BINS;
        }
	dx=(ab[1]-ab[0])/((double)(nbin));

  	if (m != SMPAR) m=SMPAR;

	/* -=-=-=-=-=-=-=-=-=- */
        /* Do the next step    */       
        /* -=-=-=-=-=-=-=-=-=- */

	if (INTERACTIVE==0) {	
		/* Forward in Non-Interactive */
	    if (DIRECTION==1) {
		step++;
	    	for (k=0;k<pp;k++) lambda[k]=lambda_store[(step-1)*pp+k];
	    } else {
		for (k=0;k<pp;k++) lambda[k]=-lambda_store[(step-1)*pp+k];
		step--;
	    }
	    rotated(X,n,p,lambda,dt,wtmatrix,pp,DIRECTION);
	    fl_set_counter_value(dgtform->frame,((double)(step)));
            for (k=1;k<=p;k++) wts[k-1]=wtmatrix[(k-1)*p];
	    if (BRUSH==0) {
                bin1(X,&n,ab,&nbin,nc,&nskip);
                ash1(&m,nc,&nbin,ab,kopt,xx,f,w,&ier);
                maxf=0.0;
                for (k=1;k<=nbin;k++) {
                    if (f[k-1]>maxf) maxf=f[k-1];
                }
                for (k=1;k<=nbin;k++) f[k-1]/=maxf;
		startquad();
                contour1d(v,xx,f,&nbin,&nv,BRUSH,0);
                if (DRAWWTS==1) addwts1d(p,wts,-0.4);
                if (DRAWPTS==1) addpts1d(X,J,&n,BRUSH,G);
		endquad();
	    } else {
                maxf=0.0;
                loc=0;
                for (k=1;k<=ng;k++) {
                    if (k>1) loc+=Group_Numbers[k-2];
                    newloc=X;
                    newloc+=loc;
                    number=Group_Numbers;
                    number+=k-1;
                    bin1(newloc,number,ab,&nbin,nc,&nskip);
                    ash1(&m,nc,&nbin,ab,kopt,xx,f,w,&ier);
                    for (l=1;l<=nbin;l++) {
			if (f[l-1]>maxf) maxf=f[l-1];
			ff[l-1][k-1]=f[l-1];
		    }
                }
		startquad();
		for (k=1;k<=ng;k++) {
                    for (l=1;l<=nbin;l++)
                        f[l-1]=ff[l-1][k-1]/maxf;
		    contour1d(v,xx,f,&nbin,&nv,BRUSH,k);
                }
                if (DRAWWTS==1)
                    addwts1d(p,wts,(-((double)(ng))*0.2-0.2));
                if (DRAWPTS==1) addpts1d(X,J,&n,BRUSH,G);
		endquad();
            }
	} 
	while (INTERACTIVE==1) {
	    USERACT=0;
	    obj=fl_check_forms();
	    NEWSTEP=fl_get_counter_value(dgtform->frame);
		if (NEWSTEP != step) USERACT=1;
            BINS=((int) fl_get_counter_value(dgtform->binsX));
		if (BINS !=nbin) {
		    nbin=BINS;
		    USERACT=1;	
    		    dx=(ab[1]-ab[0])/((double)(nbin));
		}
            SMPAR=((int) fl_get_counter_value(dgtform->SmParX));
		if (m != SMPAR) {
		    m=SMPAR;
		    USERACT=1;
		}
            QUIT=fl_get_button(dgtform->quit);
	 	if (QUIT==1) exit(1);

            INTERACTIVE=fl_get_button(dgtform->frame_by_frame);
	    Temp_Int=fl_get_button(dgtform->weights);
            	if (DRAWWTS != Temp_Int) {
		    DRAWWTS=Temp_Int;
		    USERACT=1;
	    	}
            Temp_Int=fl_get_button(dgtform->drawpts);
		if (DRAWPTS != Temp_Int) {
                    DRAWPTS=Temp_Int;
                    USERACT=1;
                }	
            Temp_Int=fl_get_button(dgtform->brushX);
		if (BRUSH != Temp_Int) {      
                    BRUSH=Temp_Int;  
                    USERACT=1; 
                }

	    if (USERACT==1) {
		STEPDIFF=NEWSTEP-step;
		if (STEPDIFF>0)  
		    while (step<NEWSTEP) {
			step++;
			DIRECTION=1;
                        for (k=0;k<pp;k++)lambda[k]=lambda_store[(step-1)*pp+k];
                        rotated(X,n,p,lambda,dt,wtmatrix,pp,DIRECTION);
		    }
		else if (STEPDIFF<0) 
		    while (step>NEWSTEP) {
			DIRECTION=0;
		      for (k=0;k<pp;k++)lambda[k]=-lambda_store[(step-1)*pp+k];
                        rotated(X,n,p,lambda,dt,wtmatrix,pp,DIRECTION);
			step--;
		    }
		step=NEWSTEP;
		fl_set_counter_value(dgtform->frame,((double)(step)));
            	for (k=1;k<=p;k++) wts[k-1]=wtmatrix[(k-1)*p];
            	if (BRUSH==0) {
                    bin1(X,&n,ab,&nbin,nc,&nskip);
                    ash1(&m,nc,&nbin,ab,kopt,xx,f,w,&ier);
                    maxf=0.0;
                    for (k=1;k<=nbin;k++) if (f[k-1]>maxf) maxf=f[k-1];
                    for (k=1;k<=nbin;k++) f[k-1]/=(maxf/2.0);
		    startquad();
                    contour1d(v,xx,f,&nbin,&nv,BRUSH,0);
                    if (DRAWWTS==1) addwts1d(p,wts,-0.4);
                    if (DRAWPTS==1) addpts1d(X,J,&n,BRUSH,G);
		    endquad();
		} else {
		    maxf=0.0;
                    loc=0;
                    for (k=1;k<=ng;k++) {
                    	if (k>1) loc+=Group_Numbers[k-2];
                    	newloc=X;
                    	newloc+=loc;
                    	number=Group_Numbers;
                    	number+=k-1;
                    	bin1(newloc,number,ab,&nbin,nc,&nskip);
                    	ash1(&m,nc,&nbin,ab,kopt,xx,f,w,&ier);
                    	for (l=1;l<=nbin;l++) {
                            if (f[l-1]>maxf) maxf=f[l-1];
                            ff[l-1][k-1]=f[l-1];
                        }
                    } 
                    startquad();
                    for (k=1;k<=ng;k++) {
                    	for (l=1;l<=nbin;l++)
                            f[l-1]=ff[l-1][k-1]/maxf;
                    	contour1d(v,xx,f,&nbin,&nv,BRUSH,k);
                    }
                    if (DRAWWTS==1) addwts1d(p,wts,(-((double)(ng))*0.2-0.2));
                    if (DRAWPTS==1) addpts1d(X,J,&n,BRUSH,G);
                    endquad();
            	}
	    }
	}	
    }
    free(nc);
    free(X);
    free(f);
    free(w);
    free(xx);
    free(yy);
    free(v);
}

void rotated(X, n, p, lambda, dt, wtmatrix, pp, dir) 
    double *X, *lambda, dt, *wtmatrix;
    int n,p,pp,dir;
{
    int i,j,k,l,ll,pcount;
    static double *alpha = NULL;
    static double *R, *Rij, *prod;
    double la;

    void multmat();

    if ( alpha == NULL ) {
        prod=(double *) calloc((n*p),sizeof(double));
        R=(double *) calloc((p*p),sizeof(double));
        Rij=(double *) calloc((p*p),sizeof(double));
        alpha=(double *) calloc(pp,sizeof(double));
    }

    for (i=1;i<=p;i++) for (j=1;j<=p;j++) {
	if (i==j) R[(j-1)*p+i-1]=1.0;
	else R[(j-1)*p+i-1]=0.0;
	if (i==j) Rij[(j-1)*p+i-1]=1.0;
        else Rij[(j-1)*p+i-1]=0.0;
    }

    for (i=1;i<=pp;i++) { 
	alpha[i-1]=dt*lambda[i-1];  /* dt * lambda[i-1] */
    }
 
    pcount=-1;
    for (j=1;j<=(p-1);j++) for (k=(j+1);k<=p;k++) {
	pcount++;

	Rij[(j-1)*p+j-1]=Rij[(k-1)*p+k-1]=cos(alpha[pcount]);
	Rij[(k-1)*p+j-1]=-sin(alpha[pcount]);
	Rij[(j-1)*p+k-1]=sin(alpha[pcount]);

        if (dir==1) multmat(R,p,p,Rij,p,p,prod); 	
	else multmat(Rij,p,p,R,p,p,prod);
	for (l=0;l<(p*p);l++) R[l]=prod[l];

	Rij[(j-1)*p+j-1]=Rij[(k-1)*p+k-1]=1.0;
        Rij[(k-1)*p+j-1]=0.0;
        Rij[(j-1)*p+k-1]=0.0;
    } 
    multmat(X,n,p,R,p,p,prod);
    for (l=0;l<(n*p);l++) X[l]=prod[l];

    multmat(wtmatrix,p,p,R,p,p,prod);
    for (l=0;l<(p*p);l++) wtmatrix[l]=prod[l];
}

void newlambda(lambda,p,idum)
    double *lambda;
    int p;
    long *idum;
{
    int i;
    double sl;

    sl=0.0;
    for (i=1;i<=p;i++) {
        lambda[i-1]=pi*(ran1(idum)*2.0-1.0);
    }
}

void multmat(X,a,b,Y,c,d,Z) 

    double *X, *Y, *Z;
    int a,b,c,d;
{
    int i,j,k;
    double sum;

    if (b!=c) exit(1);

    for (i=1;i<=a;i++) for (j=1;j<=d;j++) {
  	sum=0.0;
	for (k=1;k<=b;k++) sum+=X[(k-1)*a+i-1]*Y[(j-1)*c+k-1];
	Z[(j-1)*a+i-1]=sum;	
    }
}
void startquad()
{
    printf("(read geometry { define foo \n");
    printf("CQUAD\n");
}

void endquad()
{
    printf(" }) \n");
    fflush(stdout);
}
